
import { RepresentationTypeIdentifier } from '../Data/RepresentationType.js';
import TemplateType from '../Resource/Template/TemplateType.js';
import Warehouse from '../Resource/Warehouse.js';
import * as fs from 'fs';

export default class TemplateGenerator {
  
  static toLiteral(input) {
    if (input == null) return "null";

    let literal = "";

    literal += "\"";

    input.runes.forEach((code) => {
      var c = String.fromCharCode(code);
      switch (c) {
        case '"':
          literal += "\\\"";
          break;
        case '\\':
          literal += "\\\\";
          break;
        case '\0':
          literal += "\\0";
          break;
        case '\a':
          literal += "\\a";
          break;
        case '\b':
          literal += "\\b";
          break;
        case '\f':
          literal += "\\f";
          break;
        case '\n':
          literal += "\\n";
          break;
        case '\r':
          literal += "\\r";
          break;
        case '\t':
          literal += "\\t";
          break;
        case '\v':
          literal += "\\v";
          break;
        default:
          literal += c;
          break;
      }
    });

    literal += "\"";
    return literal;
  }

  static generateRecord(template, templates) {
    let className = template.className.split('.').last;
    let rt = "";

    let parentName;

    if (template.parentId != null) {
      parentName = this._translateClassName(templates
          .find((x) =>
              (x.classId == template.parentId) &&
              (x.type == TemplateType.Record))
          .className);
      rt += `class ${className} extends ${parentName} {\r\n`;
    } else {
      rt += "class ${className} extends IRecord { \r\n";
    }

    template.properties.forEach((p) => {
      if (p.inherited) return;
      let ptTypeName = this.getTypeName(template, p.valueType, templates);
      rt += `${ptTypeName} ${p.name};\r\n\r\n`;
    });

    rt += "\r\n";

    
    // rt += "deserialize(Map<String, dynamic> value) {");

    // template.properties.forEach((p) {
    //   rt.writeln("${p.name} = value['${p.name}'];");
    // });

    rt += "}\r\n";

    // rt.writeln("Map<String, dynamic> serialize() {");
    // rt.writeln("var rt = Map<String, dynamic>();");

    // template.properties.forEach((p) {
    //   rt.writeln("rt['${p.name}'] = ${p.name};");
    // });

    // rt.writeln("return rt;");
    // rt.writeln("}");

    // add template
    var descProps = template.properties 
        .map((p) => {
      var ptTypeName = this.getTypeName(template, p.valueType, templates);
      return `Prop('${p.name}', getTypeOf<${ptTypeName}>(), ${this._escape(p.readAnnotation)}, ${this._escape(p.writeAnnotation)})`;
    }).join(', ');

    if (parentName != null)
      rt += `TemplateDescriber get template => TemplateDescriber('${template.className}', parent: ${parentName}, properties: [${descProps}], annotation: ${this.toLiteral(template.annotation)});`;
    else
      rt += `TemplateDescriber get template => TemplateDescriber('${template.className}', properties: [${descProps}], annotation: ${this.toLiteral(template.annotation)});`;

    rt += "\r\n}";

    return rt;
  }

  static _translateClassName(className) {
    var cls = className.split('.');
    var nameSpace = cls.take(cls.length - 1).join('_').toLowerCase();
    return `${nameSpace}.${cls.last}`;
  }

  static getTypeName(forTemplate, representationType, templates) {
    let name;

    if (representationType.identifier == RepresentationTypeIdentifier.TypedResource) {
      if (representationType.guid == forTemplate.classId)
        name = forTemplate.className.split('.').last;
      else
        name = this._translateClassName(templates
            .find((x) =>
                x.classId == representationType.guid &&
                (x.type == TemplateType.Resource ||
                    x.type == TemplateType.Wrapper))
            .className);
    } else if (representationType.identifier == RepresentationTypeIdentifier.TypedRecord) {
      if (representationType.guid == forTemplate.classId)
        name = forTemplate.className.split('.').last;
      else
        name = this._translateClassName(templates
            .find((x) =>
                x.classId == representationType.guid &&
                x.type == TemplateType.Record)
            .className);
    } else if (representationType.identifier == RepresentationTypeIdentifier.Enum) {
     if (representationType.guid == forTemplate.classId)
      name = forTemplate.className.split('.').last;
    else
      name = this._translateClassName(templates
          .singleWhere((x) =>
              x.classId == representationType.guid &&
              x.type == TemplateType.Enum)
          .className);
     } else if (representationType.identifier == RepresentationTypeIdentifier.TypedList)
      name = "List<" + this.getTypeName(forTemplate, representationType.subTypes[0], templates) + ">";
    else if (representationType.identifier ==
        RepresentationTypeIdentifier.TypedMap)
      name = "Map<" +
          this.getTypeName(forTemplate, representationType.subTypes[0], templates) +
          "," +
          this.getTypeName(forTemplate, representationType.subTypes[1], templates) +
          ">";
    else if (representationType.identifier ==
            RepresentationTypeIdentifier.Tuple2 ||
        representationType.identifier == RepresentationTypeIdentifier.Tuple3 ||
        representationType.identifier == RepresentationTypeIdentifier.Tuple4 ||
        representationType.identifier == RepresentationTypeIdentifier.Tuple5 ||
        representationType.identifier == RepresentationTypeIdentifier.Tuple6 ||
        representationType.identifier == RepresentationTypeIdentifier.Tuple7)
      name = "Tuple";
    else {
      switch (representationType.identifier) {
        case RepresentationTypeIdentifier.Dynamic:
          name = "dynamic";
          break;
        case RepresentationTypeIdentifier.Bool:
          name = "bool";
          break;
        case RepresentationTypeIdentifier.Char:
          name = "String";
          break;
        case RepresentationTypeIdentifier.DateTime:
          name = "DateTime";
          break;
        case RepresentationTypeIdentifier.Decimal:
          name = "double";
          break;
        case RepresentationTypeIdentifier.Float32:
          name = "double";
          break;
        case RepresentationTypeIdentifier.Float64:
          name = "double";
          break;
        case RepresentationTypeIdentifier.Int16:
          name = "int";
          break;
        case RepresentationTypeIdentifier.Int32:
          name = "int";
          break;
        case RepresentationTypeIdentifier.Int64:
          name = "int";
          break;
        case RepresentationTypeIdentifier.Int8:
          name = "int";
          break;
        case RepresentationTypeIdentifier.String:
          name = "String";
          break;
        case RepresentationTypeIdentifier.Map:
          name = "Map";
          break;
        case RepresentationTypeIdentifier.UInt16:
          name = "int";
          break;
        case RepresentationTypeIdentifier.UInt32:
          name = "int";
          break;
        case RepresentationTypeIdentifier.UInt64:
          name = "int";
          break;
        case RepresentationTypeIdentifier.UInt8:
          name = "int";
          break;
        case RepresentationTypeIdentifier.List:
          name = "List";
          break;
        case RepresentationTypeIdentifier.Resource:
          name = "IResource";
          break;
        case RepresentationTypeIdentifier.Record:
          name = "IRecord";
          break;
        default:
          name = "dynamic";
      }
    }

    return (representationType.nullable) ? name + "?" : name;
  }

  static isNullOrEmpty(v) {
    return v == null || v == "";
  }

  static async getTemplate(url, dir, username,  password, asyncSetters = true)  {

       // var fs = require('fs');

        const _urlRegex = /^(?:([^\s|:]*):\/\/([^\/]*)\/?(.*))/;

        if (!_urlRegex.test(url)) 
         throw Error(`Invalid IIP URL '${url}'`);

      let path = url.split(_urlRegex);

      let con = await Warehouse.get(path[1] + "://" + path[2], username != null
              ? {"username": username, "password": password ?? ""}
              : null);

      if (con == null) 
        throw Error("Can't connect to server");

      if (dir == null || dir == "") 
        dir = path[2].replaceAll(":", "_");

      let templates = await con.getLinkTemplates(path[3]);

      // no longer needed
      Warehouse.remove(con);

      let dstDir = `lib/${dir}`;


      if (!fs.existsSync(dstDir)){
        fs.mkdirSync(dstDir, { recursive: true });
      }

      var makeImports = (skipTemplate) => {
        let imports = "";
        imports += "import 'dart:async';\r\n";
        imports += "import 'package:esiur/esiur.dart';\r\n";

        // make import names
        templates.forEach((tmp) => {
          if (tmp != skipTemplate) {
            let cls = tmp.className.split('.');
            let nameSpace = cls.take(cls.length - 1).join('_').toLowerCase();
            imports += `import '${tmp.className}.g.js' as ${nameSpace};`;
          }
        });

        imports += "\r\n";
        return imports;
      };

      // make sources
      templates.forEach((tmp) => {
        console.log(`Generating '${tmp.className}'.`);
        let filePath = `${dstDir.path}/${tmp.className}.g.js`;
        let f = File(filePath);

        var source = "";
        if (tmp.type == TemplateType.Resource) {
          source = makeImports(tmp) + this.generateClass(tmp, templates, asyncSetters);
        } else if (tmp.type == TemplateType.Record) {
          source = makeImports(tmp) + this.generateRecord(tmp, templates);
        } else if (tmp.type == TemplateType.Enum) {
          source = makeImports(tmp) + this.generateEnum(tmp, templates);
        }
        f.writeAsStringSync(source);
      });

      // generate info class
      // Warehouse.defineType<test.MyService>(
      //     () => test.MyService(),
      //     RepresentationType(RepresentationTypeIdentifier.TypedResource, false,
      //         Guid(DC.fromList([1, 2, 3]))));

      var defineCreators = templates.map((tmp) => {
        // creator
        var className = this._translateClassName(tmp.className);
        if (tmp.type == TemplateType.Resource ||
            tmp.type == TemplateType.Wrapper) {
          return `Warehouse.defineType<${className}>(() => ${className}(), RepresentationType(RepresentationTypeIdentifier.TypedResource, false, Guid.parse('${tmp.classId.toString()}')));\r\n`;
        } else if (tmp.type == TemplateType.Record) {
          return `Warehouse.defineType<${className}>(() => ${className}(), RepresentationType(RepresentationTypeIdentifier.TypedRecord, false, Guid.parse('${tmp.classId.toString()}')));\r\n`;
        } else if (tmp.type == TemplateType.Enum) {
          return `Warehouse.defineType<${className}>(() => ${className}(), RepresentationType(RepresentationTypeIdentifier.Enum, false, Guid.parse('${tmp.classId.toString()}')));\r\n`;
        }
      }).join("\r\n");

      var putTemplates = templates.map((tmp) => {
        var className = this._translateClassName(tmp.className);
        return `Warehouse.putTemplate(TypeTemplate.fromType(${className}));`;
      }).join("\r\n");

      var typesFile = makeImports(null) +
          `\r\n void init_${dir?.replaceAll('.', '_')}(){ ${defineCreators} \r\n ${putTemplates}}`;

      var f = `${dstDir.path}/init.g.dart`;
      await fs.writeFile(f, typesFile);

      return dstDir;
 
  }

  static _escape(str) {
    if (str == null)
      return "null";
    else
      return "r'$str'";
  }

  static generateEnum(template, templates) {
    var className = template.className.split('.').last;
    var rt = "";

    rt += `class ${className} extends IEnum {\r\n`;

    template.constants.forEach((c) => {
      rt += `static ${className} ${c.name} = ${className}(${c.index}, ${c.value}, '${c.name}');\r\n`;
    });

    rt += "\r\n";

    rt += `${className}([int index = 0, value, String name = '']) : super(index, value, name);`;

    // add template
    var descConsts = template.constants.map((p) => {
      var ctTypeName = this.getTypeName(template, p.valueType, templates);
      return `Const('${p.name}', getTypeOf<${ctTypeName}>(), ${p.value}, ${this._escape(p.annotation)})`;
    }).join(', ');

    rt += `TemplateDescriber get template => TemplateDescriber('${template.className}', constants: [${descConsts}], annotation: ${this.toLiteral(template.annotation)});`;

    rt += "\r\n}";

    return rt;
  }

  static generateClass(template, templates, asyncSetters = true) {
    let className = template.className.split('.').last;

    let parentName = null;

    let rt = "";

    if (template.parentId != null) {
      parentName = this._translateClassName(templates
          .find((x) =>
              (x.classId == template.parentId) &&
              (x.type == TemplateType.Resource ||
                  x.type == TemplateType.Wrapper))
          .className);
      rt += `class ${className} extends ${parentName} {\r\n`;
    } else {
      rt += `class ${className} extends DistributedResource {\r\n`;
    }

    rt += `${className}() {`;

    template.events.filter((e) => !e.inherited).forEach((e) => {
      rt += `on('${e.name}', (x) => _${e.name}Controller.add(x));`;
    });

    rt += "}";


    template.functions.filter((f) => !f.inherited).forEach((f) => {
      var rtTypeName = this.getTypeName(template, f.returnType, templates);
      var positionalArgs = f.arguments.filter((x) => !x.optional);
      var optionalArgs = f.arguments.filter((x) => x.optional);

      if (f.isStatic) {
        rt += `static AsyncReply<${rtTypeName}> ${f.name}(DistributedConnection connection`;

        if (positionalArgs.length > 0)
          rt += `, ${positionalArgs.map((a) => this.getTypeName(template, a.type, templates) + " " + a.name).join(',')}`;

        if (optionalArgs.length > 0) {
          rt += `, [${optionalArgs.map((a) => this.getTypeName(template, a.type.toNullable(), templates) + " " + a.name).join(',')}]`;
        }
      } else {
        rt += `AsyncReply<$rtTypeName> ${f.name}(`;

        if (positionalArgs.length > 0)
          rt += `${positionalArgs.map((a) => this.getTypeName(template, a.type, templates) + " " + a.name).join(',')}`;

        if (optionalArgs.length > 0) {
          if (positionalArgs.length > 0) rt.write(",");
          rt += `[${optionalArgs.map((a) => this.getTypeName(template, a.type.toNullable(), templates) + " " + a.name).join(',')}]`;
        }
      }

      rt += ") {\r\n";

      rt += `var args = <UInt8, dynamic>{${positionalArgs.map((e) => "UInt8(" + e.index.toString() + ') :' + e.name).join(',')}};\r\n`;

      optionalArgs.forEach((a) => {
        rt += `if (${a.name} != null) args[UInt8(${a.index})] = ${a.name};\r\n`;
      });

      rt += `var rt = AsyncReply<$rtTypeName>();\r\n`;
      if (f.isStatic) {
        rt += `connection.staticCall(Guid.parse('${template.classId.toString()}'), ${f.index}, args)`;
      } else {
        rt += `internal_invoke(${f.index}, args)`;
      }
      rt += `.then((x) => rt.trigger(x))\r\n`;
      rt += `.error((x) => rt.triggerError(x))\r\n`;
      rt += `.chunk((x) => rt.triggerChunk(x));\r\n`;
      rt += `return rt; }\r\n`;
    });

    template.properties.filter((p) => !p.inherited).forEach((p) => {
      let ptTypeName = this.getTypeName(template, p.valueType, templates);
      rt += `${ptTypeName} get ${p.name} { return get(${p.index}); }\r\n`;

      if (asyncSetters)
        rt += `set ${p.name}(${ptTypeName} value) { set(${p.index}, value); }\r\n`;
      else
        rt += `set ${p.name}(${ptTypeName} value) { setSync(${p.index}, value); }\r\n`;
    });

    template.events.filter((e) => !e.inherited).forEach((e) => {
      var etTypeName = this.getTypeName(template, e.argumentType, templates);

      rt += `final _${e.name}Controller = StreamController<$etTypeName>();\r\n`;
      rt += `Stream<${etTypeName}> get ${e.name} { \r\n`;
      rt += `return _${e.name}Controller.stream;\r\n`;
      rt += "}";
    });

    // add template
    var descProps = template.properties //.where((p) => !p.inherited)
        .map((p) => {
      var ptTypeName = this.getTypeName(template, p.valueType, templates);
      return `Prop('${p.name}', getTypeOf<${ptTypeName}>(), ${this._escape(p.readAnnotation)}, ${this._escape(p.writeAnnotation)})`;
    }).join(', ');

    var descFuncs = template.functions
        .map((f) => {
      var ftTypeName = this.getTypeName(template, f.returnType, templates);

      var args = f.arguments.map((a) => {
        var atTypeName = this.getTypeName(template, a.type, templates);
        return `Arg('${a.name}', getTypeOf<${atTypeName}>(), ${a.optional})`;
      }).join(', ');

      return `Func('${f.name}', getTypeOf<${ftTypeName}>(), [${args}], ${this._escape(f.annotation)})`;
    }).join(', ');

    var descEvents = template.events
        //.where((e) => !e.inherited)
        .map((e) => {
      var etTypeName = this.getTypeName(template, e.argumentType, templates);
      return `Evt('${e.name}', getTypeOf<${etTypeName}>(), ${e.listenable}, ${this._escape(e.annotation)})`;
    }).join(', ');

    if (parentName != null)
      rt += `TemplateDescriber get template => TemplateDescriber('${template.className}', parent: ${parentName}, properties: [${descProps}], functions: [${descFuncs}], events: [${descEvents}], annotation: ${this.toLiteral(template.annotation)});\r\n`;
    else
      rt += `TemplateDescriber get template => TemplateDescriber('${template.className}', properties: [${descProps}], functions: [${descFuncs}], events: [${descEvents}], annotation: ${this.toLiteral(template.annotation)});\r\n`;

    rt += "\r\n}";

    return rt;
  }
}
